from . import modules
from . import utils
from .modules.models import hourglass_diffusion_transformer
from .modules.models.hourglass_diffusion_transformer import HourglassDiffusionTransformer
from .modules.models import hdit
from .modules.models.hdit import HDiT


"""k-diffusion transformer diffusion models, version 2."""
from typing import Iterable, Optional, Callable, Dict, overload
from functools import reduce
from typing import Union
from einops import rearrange
import numpy as np
import torch
from torch import nn
import torch._dynamo
from ....utils.args_context import ArgsContext
from ... import flags
from . import layers
from ...attention.axial_rope import make_axial_pos
from ...modules import apply_wd, zero_init, RMSNorm, LinearGEGLU, Linear, TokenMerge, TokenSplit, TokenSplitWithoutSkip, Identity
from ...attention import NeighborhoodTransformerLayer, GlobalTransformerLayer


if flags.get_use_compile():
    torch._dynamo.config.cache_size_limit = max(64, torch._dynamo.config.cache_size_limit)
    torch._dynamo.config.suppress_errors = True


# Helpers
def downscale_pos(pos):
    pos = rearrange(pos, "... (h nh) (w nw) e -> ... h w (nh nw) e", nh=2, nw=2)
    return torch.mean(pos, dim=-2)


def filter_params(function, module):
    for param in module.parameters():
        tags = getattr(param, "_tags", set())
        if function(tags):
            yield param


# Rotary position embeddings
@flags.compile_wrap
def apply_rotary_emb(x, theta, conj=False):
    out_dtype = x.dtype
    dtype = reduce(torch.promote_types, (x.dtype, theta.dtype, torch.float32))
    d = theta.shape[-1]
    assert d * 2 <= x.shape[-1]
    x1, x2, x3 = x[..., :d], x[..., d : d * 2], x[..., d * 2 :]
    x1, x2, theta = x1.to(dtype), x2.to(dtype), theta.to(dtype)
    cos, sin = torch.cos(theta), torch.sin(theta)
    sin = -sin if conj else sin
    y1 = x1 * cos - x2 * sin
    y2 = x2 * cos + x1 * sin
    y1, y2 = y1.to(out_dtype), y2.to(out_dtype)
    return torch.cat((y1, y2, x3), dim=-1)


class Level(nn.ModuleList):
    def forward(self, x, *args, **kwargs):
        for layer in self:
            x = layer(x, *args, **kwargs)
        return x


# Mapping network
class MappingFeedForwardBlock(nn.Module):
    def __init__(self, d_model, d_ff, dropout=0.0):
        super().__init__()
        self.norm = RMSNorm(d_model)
        self.up_proj = apply_wd(LinearGEGLU(d_model, d_ff, bias=False))
        self.dropout = nn.Dropout(dropout)
        self.down_proj = apply_wd(zero_init(Linear(d_ff, d_model, bias=False)))

    def forward(self, x):
        skip = x
        x = self.norm(x)
        x = self.up_proj(x)
        x = self.dropout(x)
        x = self.down_proj(x)
        return x + skip


class MappingNetwork(nn.Module):
    def __init__(self, n_layers, d_model, d_ff, dropout=0.0):
        super().__init__()
        self.in_norm = RMSNorm(d_model)
        self.blocks = nn.ModuleList([MappingFeedForwardBlock(d_model, d_ff, dropout=dropout) for _ in range(n_layers)])
        self.out_norm = RMSNorm(d_model)

    def forward(self, x):
        x = self.in_norm(x)
        for block in self.blocks:
            x = block(x)
        x = self.out_norm(x)
        return x


class SelectiveNeighborhoodTransformerLayer(NeighborhoodTransformerLayer):
    def __init__(self, d_model, d_ff, d_head, cond_features, kernel_size, dropout=0, cond_index=None):
        super().__init__(d_model, d_ff, d_head, cond_features, kernel_size, dropout)
        self.cond_index = cond_index
        
    def forward(self, x, pos, cond, *args):
        if self.cond_index is None:
            return super().forward(x, pos, cond)
        return super().forward(x, pos, cond, args[self.cond_index])


class SelectiveGlobalTransformerLayer(GlobalTransformerLayer):
    def __init__(self, d_model, d_ff, d_head, cond_features, dropout=0, cond_index=None):
        super().__init__(d_model, d_ff, d_head, cond_features, dropout)
        self.cond_index = cond_index
    
    def forward(self, x, pos, cond, *args):
        if self.cond_index is None:
            return super().forward(x, pos, cond)
        return super().forward(x, pos, cond, args[self.cond_index])


default_layer_block_builders = {
    0: lambda context: SelectiveNeighborhoodTransformerLayer(context.width, context.width, 64, context.mapping_width, 7, 0., cond_index=None),
    1: lambda context: SelectiveNeighborhoodTransformerLayer(context.width, context.width, 64, context.mapping_width, 7, 0., cond_index=None),
    2: lambda context: SelectiveGlobalTransformerLayer(context.width, context.width, 64, context.mapping_width, 0., cond_index=None),
    3: lambda context: SelectiveNeighborhoodTransformerLayer(context.width, context.width, 64, context.mapping_width, 7, 0., cond_index=None),
    4: lambda context: SelectiveNeighborhoodTransformerLayer(context.width, context.width, 64, context.mapping_width, 7, 0., cond_index=None),
}


class HourglassDiffusionTransformer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        local_condition_channels: Optional[int] = None,
        patch_size: int = [4, 4],
        widths: Iterable[int] = [128, 256],
        middle_width: int = 512,
        depths: Iterable[int] = [2, 2],
        middle_depth: int = 4,
        layer_block_builders: Dict[int, Callable[[ArgsContext], nn.Module]] = default_layer_block_builders,
        mapping_width: int = 256,
        mapping_depth: int = 2,
        mapping_feed_forward_dim: Optional[int] = None,
        mapping_dropout: float = 0.,
        is_layer_condition_module_prior: bool = True,
        layer_condition_initialization_modules: Union[nn.Module, Iterable[nn.Module], None] = None,
        layer_condition_guidance_modules: Union[nn.Module, Iterable[nn.Module], None] = None,
        layer_condition_guidance_dims: Union[int, Iterable[Optional[int]], None] = None,
        guidance_modules: Union[nn.Module, Iterable[nn.Module], None] = None,
        guidance_dims: Union[int, Iterable[Optional[int]], None] = None,
    ):
        super().__init__()
        
        self.layer_block_builders = layer_block_builders
        self.mapping_width = mapping_width
        
        if local_condition_channels is None:
            self.local_condition_channels = 0
        else:
            self.local_condition_channels = local_condition_channels

        self.patch_in = TokenMerge(in_channels + self.local_condition_channels, widths[0], patch_size)

        self.time_embedding = layers.FourierFeatures(1, mapping_width)
        self.time_projection = Linear(mapping_width, mapping_width, bias=False)
        
        if layer_condition_initialization_modules is None:
            layer_condition_initialization_modules = []
        elif not isinstance(layer_condition_initialization_modules, Iterable):
            layer_condition_initialization_modules = [layer_condition_initialization_modules]
        if layer_condition_guidance_modules is None:
            layer_condition_guidance_modules = []
        elif not isinstance(layer_condition_guidance_modules, Iterable):
            layer_condition_guidance_modules = [layer_condition_guidance_modules]
        if layer_condition_guidance_dims is None or not isinstance(layer_condition_guidance_dims, Iterable):
            layer_condition_guidance_dims = [layer_condition_guidance_dims for _ in range(len(layer_condition_guidance_modules))]
        else:
            layer_condition_guidance_dims += [None for _ in range(len(layer_condition_guidance_modules) - len(layer_condition_guidance_dims))]
        if guidance_modules is None:
            guidance_modules = []
        elif not isinstance(guidance_modules, Iterable):
            guidance_modules = [guidance_modules]
        if guidance_dims is None or not isinstance(guidance_dims, Iterable):
            guidance_dims = [guidance_dims for _ in range(len(guidance_modules))]
        else:
            guidance_dims += [None for _ in range(len(guidance_modules) - len(guidance_dims))]
        self.layer_condition_initialization_modules = nn.ModuleList(layer_condition_initialization_modules)
        self.layer_condition_guidance_modules = nn.ModuleList(layer_condition_guidance_modules)
        self.guidance_modules = nn.ModuleList(guidance_modules)
        
        self.guidance_modules_projections = nn.ModuleList([
            Linear(mapping_width if guidance_dim is None else guidance_dim, mapping_width, bias=False) for guidance_dim in guidance_dims
        ])
        self.is_layer_condition_module_prior = is_layer_condition_module_prior
        self.layer_condition_guidance_module_projections = nn.ModuleList([
            Linear(mapping_width if layer_cond_guidance_dim is None else layer_cond_guidance_dim, mapping_width, bias=False) for layer_cond_guidance_dim in layer_condition_guidance_dims
        ])
        
        self.mapping = MappingNetwork(mapping_depth, mapping_width, mapping_feed_forward_dim, dropout=mapping_dropout)

        self.down_levels, self.up_levels = nn.ModuleList(), nn.ModuleList()
        num_half_levels = len(widths)
        self.num_total_levels = 2 * num_half_levels + 1
        for i, (width, depth) in enumerate(zip(widths, depths)):
            self.down_levels.append(Level([self.build_layer_block(width, j, i) for j in range(depth)]))
            self.up_levels.append(Level([self.build_layer_block(width, j + depth, self.num_total_levels - 1 - i) for j in range(depth)]))
        self.middle_level = Level([self.build_layer_block(middle_width, j, num_half_levels) for j in range(middle_depth)])
        
        no_middle_widths = widths
        no_first_widths = np.concatenate((widths[1:], [middle_width]))

        self.merges = nn.ModuleList([TokenMerge(width_1, width_2) for width_1, width_2 in zip(no_middle_widths, no_first_widths)])
        self.splits = nn.ModuleList([TokenSplit(width_2, width_1) for width_1, width_2 in zip(no_middle_widths, no_first_widths)])

        self.out_norm = RMSNorm(widths[0])
        self.patch_out = TokenSplitWithoutSkip(widths[0], out_channels, patch_size)
        nn.init.zeros_(self.patch_out.proj.weight)
        
        del self.layer_block_builders
    
    def build_layer_block(self, width: int, layer_index: int, level_index: int):
        args_context = ArgsContext(
            width=width,
            mapping_width=self.mapping_width,
            layer_index=layer_index,
            level_index=level_index
        )
        if level_index in self.layer_block_builders.keys():
            return self.layer_block_builders[level_index](args_context)
        else:
            return Identity()

    def layer_block_parameters(self, time, *args):
        layer_cond_args_length = len(args) - len(self.guidance_modules)
        # Timestep embedding
        time_embedded = self.time_embedding(time[..., None])
        block_mapping = self.time_projection(time_embedded)
        
        guidance_embedded = [guidance_module(args[layer_cond_args_length + i]) for i, guidance_module in enumerate(self.guidance_modules)]
        for i, guidance_emb in enumerate(guidance_embedded):
            block_mapping += self.guidance_modules_projections[i](guidance_emb)
        
        layer_cond_args = list(args[:layer_cond_args_length])
        layer_cond_initialized = [init_module(layer_cond_args[i]) for i, init_module in enumerate(self.layer_condition_initialization_modules)]
        layer_cond_initialized += layer_cond_args[len(layer_cond_initialized):]
        if not self.is_layer_condition_module_prior:
            layer_cond_guidance_embedded = [layer_cond_guidance_module(layer_cond_initialized[i]) for i, layer_cond_guidance_module in enumerate(self.layer_condition_guidance_modules)]
            for i, layer_cond_guidance_modules_proj in enumerate(self.layer_condition_guidance_module_projections):
                block_mapping += layer_cond_guidance_modules_proj(layer_cond_guidance_embedded[i])
        
        return block_mapping, layer_cond_initialized
    
    def add_residual(self, h, residual: Optional[torch.Tensor]):
        if residual is None:
            return h
        else:
            return h + residual.to(h.dtype)
    
    def prepare_additional_residuals(self, additional_residuals: Optional[Iterable[torch.Tensor]]):
        all_levels_length = self.num_total_levels
        additional_residuals_list = additional_residuals
        additional_residuals = np.empty(all_levels_length, dtype=object)
        if additional_residuals_list is not None:
            additional_residuals_list = np.fromiter(additional_residuals_list, dtype=object)
            additional_residuals_length = len(additional_residuals_list)
            if additional_residuals_length < all_levels_length:
                additional_residuals[:additional_residuals_length] = additional_residuals_list
            else:
                additional_residuals[:] = additional_residuals_list[:additional_residuals_length]
        return additional_residuals
    
    def split_additional_residuals(self, additional_residuals):
        num_half_levels = len(self.down_levels)
        return additional_residuals[:num_half_levels], additional_residuals[num_half_levels:-num_half_levels], additional_residuals[-num_half_levels:]
    
    def split_skip_residuals(self, skip_residuals):
        num_half_levels = len(self.down_levels)
        if isinstance(skip_residuals, Iterable):
            if len(skip_residuals) > num_half_levels:
                return skip_residuals[num_half_levels:], skip_residuals[:num_half_levels]
            else:
                return None, skip_residuals
        else:
            return None, None
    
    def pop_first_residual(self, residuals: np.ndarray):
        first_item = residuals[0]
        if len(residuals) > 1:
            other_items = residuals[1:]
        else:
            other_items = residuals[0:0]
        return first_item, other_items
    
    def concat_channels(self, x: torch.Tensor, *args):
        if self.local_condition_channels > 0:
            x = torch.cat([x, args[0]], dim=1)
            args = args[1:]
        return x, args
    
    def patch_positionally(self, x: torch.Tensor):
        # Patching
        x = x.movedim(-3, -1)
        x = self.patch_in(x)
        # TODO: pixel aspect ratio for nonsquare patches
        pos = make_axial_pos(x.shape[-3], x.shape[-2], device=x.device).view(x.shape[-3], x.shape[-2], 2)
        return x, pos
    
    def make_block_mapping(self, time: torch.Tensor, *args):
        # Mapping Network
        block_mapping, layer_cond = self.layer_block_parameters(time, *args)
        block_mapping = self.mapping(block_mapping)
        return block_mapping, layer_cond
    
    def down_sample(self, x, additional_residuals, position, block_mapping, layer_cond):
        skips, positions = [], []
        for down_level, merge in zip(self.down_levels, self.merges):
            first_item, additional_residuals = self.pop_first_residual(additional_residuals)
            x = self.add_residual(down_level(x, position, block_mapping, *layer_cond), first_item)
            skips.append(x)
            positions.append(position)
            x = merge(x)
            position = downscale_pos(position)
        return x, position, skips, positions
    
    def middle_sample(self, x, additional_residuals, pos, block_mapping, layer_cond):
        first_item, additional_residuals = self.pop_first_residual(additional_residuals)
        x = self.add_residual(self.middle_level(x, pos, block_mapping, *layer_cond), first_item)
        return x
    
    def middle_skip_connect(self, x, skip_residuals):
        if isinstance(skip_residuals, Iterable):
            skip_item = skip_residuals.pop()
            x = self.add_residual(x, skip_item)
        return x
    
    def up_sample(self, x, additional_residuals, skips, poses, block_mapping, layer_cond, *, skip_residuals):
        if isinstance(skip_residuals, Iterable):
            def split_skip(split, x, skip):
                skip_item = skip_residuals.pop()
                return split(x, skip + skip_item)
        else:
            def split_skip(split, x, skip):
                return split(x, skip)
                
        for up_level, split, skip, pos in reversed(list(zip(self.up_levels, self.splits, skips, poses))):
            first_item, additional_residuals = self.pop_first_residual(additional_residuals)
            x = split_skip(split, x, skip)
            x = self.add_residual(up_level(x, pos, block_mapping, *layer_cond), first_item)
        return x
    
    def unpatch(self, x: torch.Tensor):
        # Unpatching
        x = self.out_norm(x)
        x = self.patch_out(x)
        x = x.movedim(-1, -3)
        return x

    @overload
    def forward(self, x, t, guidance, *args, additional_residuals: Optional[Iterable[torch.Tensor]] = None, skip_residuals: Optional[Iterable[torch.Tensor]] = None):
        ...
    
    @overload
    def forward(self, x, t, layer_cond, guidance, *args, additional_residuals: Optional[Iterable[torch.Tensor]] = None, skip_residuals: Optional[Iterable[torch.Tensor]] = None):
        ...

    @overload
    def forward(self, x, t, concatenation, guidance, *args, additional_residuals: Optional[Iterable[torch.Tensor]] = None, skip_residuals: Optional[Iterable[torch.Tensor]] = None):
        ...
    
    @overload
    def forward(self, x, t, concatenation, layer_cond, guidance, *args, additional_residuals: Optional[Iterable[torch.Tensor]] = None, skip_residuals: Optional[Iterable[torch.Tensor]] = None):
        ...

    def forward(self, x: torch.Tensor, timestep, *args, additional_residuals: Optional[Iterable] = None, skip_residuals: Optional[Iterable[torch.Tensor]] = None):
        additional_residuals = self.prepare_additional_residuals(additional_residuals)
        down_additional_residuals, middle_additional_residuals, up_additional_residuals = self.split_additional_residuals(additional_residuals)
        middle_skip_residuals, up_skip_residuals = self.split_skip_residuals(skip_residuals)
        
        x, args = self.concat_channels(x, *args)
        x, pos = self.patch_positionally(x)
        
        block_mapping, layer_cond = self.make_block_mapping(timestep, *args)

        # Hourglass Transformer
        x, pos, skips, poses = self.down_sample(x, down_additional_residuals, pos, block_mapping, layer_cond)
        x = self.middle_sample(x, middle_additional_residuals, pos, block_mapping, layer_cond)
        x = self.middle_skip_connect(x, middle_skip_residuals)
        x = self.up_sample(x, up_additional_residuals, skips, poses, block_mapping, layer_cond, skip_residuals=up_skip_residuals)

        x = self.unpatch(x)
        return x


import math
import torch
import torch._dynamo
from torch import nn
from ...modules import flags


if flags.get_use_compile():
    torch._dynamo.config.suppress_errors = True


def rotate_half(x):
    x1, x2 = x[..., 0::2], x[..., 1::2]
    x = torch.stack((-x2, x1), dim=-1)
    *shape, d, r = x.shape
    return x.view(*shape, d * r)


@flags.compile_wrap
def apply_rotary_emb(freqs, t, start_index=0, scale=1.0):
    freqs = freqs.to(t)
    rot_dim = freqs.shape[-1]
    end_index = start_index + rot_dim
    assert rot_dim <= t.shape[-1], f"feature dimension {t.shape[-1]} is not of sufficient size to rotate in all the positions {rot_dim}"
    t_left, t, t_right = t[..., :start_index], t[..., start_index:end_index], t[..., end_index:]
    t = (t * freqs.cos() * scale) + (rotate_half(t) * freqs.sin() * scale)
    return torch.cat((t_left, t, t_right), dim=-1)


def centers(start, stop, num, dtype=None, device=None):
    edges = torch.linspace(start, stop, num + 1, dtype=dtype, device=device)
    return (edges[:-1] + edges[1:]) / 2


def make_grid(h_pos, w_pos):
    grid = torch.stack(torch.meshgrid(h_pos, w_pos, indexing='ij'), dim=-1)
    h, w, d = grid.shape
    return grid.view(h * w, d)


def bounding_box(h, w, pixel_aspect_ratio=1.0):
    # Adjusted dimensions
    w_adj = w
    h_adj = h * pixel_aspect_ratio

    # Adjusted aspect ratio
    ar_adj = w_adj / h_adj

    # Determine bounding box based on the adjusted aspect ratio
    y_min, y_max, x_min, x_max = -1.0, 1.0, -1.0, 1.0
    if ar_adj > 1:
        y_min, y_max = -1 / ar_adj, 1 / ar_adj
    elif ar_adj < 1:
        x_min, x_max = -ar_adj, ar_adj

    return y_min, y_max, x_min, x_max


def make_axial_pos(h, w, pixel_aspect_ratio=1.0, align_corners=False, dtype=None, device=None):
    y_min, y_max, x_min, x_max = bounding_box(h, w, pixel_aspect_ratio)
    if align_corners:
        h_pos = torch.linspace(y_min, y_max, h, dtype=dtype, device=device)
        w_pos = torch.linspace(x_min, x_max, w, dtype=dtype, device=device)
    else:
        h_pos = centers(y_min, y_max, h, dtype=dtype, device=device)
        w_pos = centers(x_min, x_max, w, dtype=dtype, device=device)
    return make_grid(h_pos, w_pos)


def freqs_pixel(max_freq=10.0):
    def init(shape):
        freqs = torch.linspace(1.0, max_freq / 2, shape[-1]) * math.pi
        return freqs.log().expand(shape)
    return init


def freqs_pixel_log(max_freq=10.0):
    def init(shape):
        log_min = math.log(math.pi)
        log_max = math.log(max_freq * math.pi / 2)
        return torch.linspace(log_min, log_max, shape[-1]).expand(shape)
    return init


class AxialRoPE(nn.Module):
    def __init__(self, dim, n_heads, start_index=0, freqs_init=freqs_pixel_log(max_freq=10.0)):
        super().__init__()
        self.n_heads = n_heads
        self.start_index = start_index
        log_freqs = freqs_init((n_heads, dim // 4))
        self.freqs_h = nn.Parameter(log_freqs.clone())
        self.freqs_w = nn.Parameter(log_freqs.clone())

    def extra_repr(self):
        dim = (self.freqs_h.shape[-1] + self.freqs_w.shape[-1]) * 2
        return f"dim={dim}, n_heads={self.n_heads}, start_index={self.start_index}"

    def get_freqs(self, pos):
        if pos.shape[-1] != 2:
            raise ValueError("input shape must be (..., 2)")
        freqs_h = pos[..., None, None, 0] * self.freqs_h.exp()
        freqs_w = pos[..., None, None, 1] * self.freqs_w.exp()
        freqs = torch.cat((freqs_h, freqs_w), dim=-1).repeat_interleave(2, dim=-1)
        return freqs.transpose(-2, -3)

    def forward(self, x, pos):
        freqs = self.get_freqs(pos)
        return apply_rotary_emb(freqs, x, self.start_index)



import math
from functools import reduce
import torch
from torch import nn
from torch.nn import init
from torch.nn import functional as F
from einops import rearrange
from . import flops, flags


class Swish(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


class ContinuityEmbedding(nn.Module):
    def __init__(self, total, d_model, dim):
        assert d_model % 2 == 0
        super().__init__()
        emb = torch.arange(0, d_model, step=2) / d_model * math.log(10000)
        emb = torch.exp(-emb)
        pos = torch.arange(total).float()
        emb = pos[:, None] * emb[None, :]
        assert list(emb.shape) == [total, d_model // 2]
        emb = torch.stack([torch.sin(emb), torch.cos(emb)], dim=-1)
        assert list(emb.shape) == [total, d_model // 2, 2]
        emb = emb.view(total, d_model)

        self.embedding = nn.Sequential(
            nn.Embedding.from_pretrained(emb),
            nn.Linear(d_model, dim),
            Swish(),
            nn.Linear(dim, dim),
        )
        self.initialize()

    def initialize(self):
        for module in self.modules():
            if isinstance(module, nn.Linear):
                init.xavier_uniform_(module.weight)
                init.zeros_(module.bias)

    def forward(self, x):
        emb = self.embedding(x)
        return emb


class DiscretenessEmbedding(nn.Module):
    def __init__(self, total, d_model, dim):
        assert d_model % 2 == 0
        super().__init__()
        self.embedding = nn.Sequential(
            nn.Embedding(num_embeddings=total + 1, embedding_dim=d_model, padding_idx=0),
            nn.Linear(d_model, dim),
            Swish(),
            nn.Linear(dim, dim),
        )

    def forward(self, x):
        emb = self.embedding(x)
        return emb


def apply_wd(module):
    for name, param in module.named_parameters():
        if name.endswith("weight"):
            tag_param(param, "wd")
    return module


def zero_init(layer):
    nn.init.zeros_(layer.weight)
    if layer.bias is not None:
        nn.init.zeros_(layer.bias)
    return layer


# Param tags
def tag_param(param, tag):
    if not hasattr(param, "_tags"):
        param._tags = set([tag])
    else:
        param._tags.add(tag)
    return param


def tag_module(module, tag):
    for param in module.parameters():
        tag_param(param, tag)
    return module


@flags.compile_wrap
def rms_norm(x, scale, eps):
    dtype = reduce(torch.promote_types, (x.dtype, scale.dtype, x.dtype))
    mean_sq = torch.mean(x.to(dtype) ** 2, dim=-1, keepdim=True)
    scale = scale.to(dtype) * torch.rsqrt(mean_sq + eps)
    return x * scale.to(x.dtype)


# Kernels
@flags.compile_wrap
def linear_geglu(x, weight, bias=None):
    x = x @ weight.mT
    if bias is not None:
        x = x + bias
    x, gate = x.chunk(2, dim=-1)
    return x * F.gelu(gate)


@flags.compile_wrap
def _apply_rotary_emb_inplace(x, theta, conj):
    dtype = reduce(torch.promote_types, (x.dtype, theta.dtype, torch.float32))
    d = theta.shape[-1]
    assert d * 2 <= x.shape[-1]
    x1, x2 = x[..., :d], x[..., d : d * 2]
    x1_, x2_, theta = x1.to(dtype), x2.to(dtype), theta.to(dtype)
    cos, sin = torch.cos(theta), torch.sin(theta)
    sin = -sin if conj else sin
    y1 = x1_ * cos - x2_ * sin
    y2 = x2_ * cos + x1_ * sin
    x1.copy_(y1)
    x2.copy_(y2)


class ApplyRotaryEmbeddingInplace(torch.autograd.Function):
    generate_vmap_rule = True

    @staticmethod
    def setup_context(ctx, inputs, output):
        x, theta, conj = inputs
        ctx.mark_dirty(x)
        ctx.save_for_backward(theta)
        ctx.save_for_forward(theta)
        ctx.conj = conj

    @staticmethod
    def forward(x, theta, conj):
        _apply_rotary_emb_inplace(x, theta, conj)
        return x

    @staticmethod
    def backward(ctx, grad_output):
        theta, = ctx.saved_tensors
        grad_output = ApplyRotaryEmbeddingInplace.apply(grad_output.clone(), theta, not ctx.conj)
        return grad_output, None, None

    @staticmethod
    def jvp(ctx, grad_input, _, __):
        theta, = ctx.saved_tensors
        return ApplyRotaryEmbeddingInplace.apply(grad_input, theta, ctx.conj)


def apply_rotary_emb_(x, theta):
    return ApplyRotaryEmbeddingInplace.apply(x, theta, False)


class AdaRMSNorm(nn.Module):
    def __init__(self, features, cond_features, eps=1e-6):
        super().__init__()
        self.eps = eps
        self.linear = apply_wd(zero_init(Linear(cond_features, features, bias=False)))
        tag_module(self.linear, "mapping")

    def extra_repr(self):
        return f"eps={self.eps},"

    def forward(self, x, cond):
        return rms_norm(x, self.linear(cond)[:, None, None, :] + 1, self.eps)


class LinearGEGLU(nn.Linear):
    def __init__(self, in_features, out_features, bias=True):
        super().__init__(in_features, out_features * 2, bias=bias)
        self.out_features = out_features

    def forward(self, x):
        flops.op(flops.op_linear, x.shape, self.weight.shape)
        return linear_geglu(x, self.weight, self.bias)


class RMSNorm(nn.Module):
    def __init__(self, shape, eps=1e-6):
        super().__init__()
        self.eps = eps
        self.scale = nn.Parameter(torch.ones(shape))

    def extra_repr(self):
        return f"shape={tuple(self.scale.shape)}, eps={self.eps}"

    def forward(self, x, *args):
        return rms_norm(x, self.scale, self.eps)


class AxialRoPE(nn.Module):
    def __init__(self, dim, n_heads):
        super().__init__()
        log_min = math.log(math.pi)
        log_max = math.log(10.0 * math.pi)
        freqs = torch.linspace(log_min, log_max, n_heads * dim // 4 + 1)[:-1].exp()
        self.register_buffer("freqs", freqs.view(dim // 4, n_heads).T.contiguous())

    def extra_repr(self):
        return f"dim={self.freqs.shape[1] * 4}, n_heads={self.freqs.shape[0]}"

    def forward(self, pos):
        theta_h = pos[..., None, 0:1] * self.freqs.to(pos.dtype)
        theta_w = pos[..., None, 1:2] * self.freqs.to(pos.dtype)
        return torch.cat((theta_h, theta_w), dim=-1)


class FeedForwardBlock(nn.Module):
    def __init__(self, d_model, d_ff, cond_features = None, dropout=0.0):
        super().__init__()
        if cond_features is None:
            self.norm = RMSNorm(d_model)
        else:
            self.norm = AdaRMSNorm(d_model, cond_features)
        self.up_proj = apply_wd(LinearGEGLU(d_model, d_ff, bias=False))
        self.dropout = nn.Dropout(dropout)
        self.down_proj = apply_wd(zero_init(Linear(d_ff, d_model, bias=False)))

    def forward(self, x, cond = None):
        skip = x
        x = self.norm(x, cond)
        x = self.up_proj(x)
        x = self.dropout(x)
        x = self.down_proj(x)
        return x + skip


class Linear(nn.Linear):
    def forward(self, x):
        flops.op(flops.op_linear, x.shape, self.weight.shape)
        return super().forward(x)


# Token merging and splitting
class TokenMerge(nn.Module):
    def __init__(self, in_features, out_features, patch_size=(2, 2)):
        super().__init__()
        self.h = patch_size[0]
        self.w = patch_size[1]
        self.proj = apply_wd(Linear(in_features * self.h * self.w, out_features, bias=False))

    def forward(self, x):
        x = rearrange(x, "... (h nh) (w nw) e -> ... h w (nh nw e)", nh=self.h, nw=self.w)
        return self.proj(x)


class TokenSplitWithoutSkip(nn.Module):
    def __init__(self, in_features, out_features, patch_size=(2, 2)):
        super().__init__()
        self.h = patch_size[0]
        self.w = patch_size[1]
        self.proj = apply_wd(Linear(in_features, out_features * self.h * self.w, bias=False))

    def forward(self, x):
        x = self.proj(x)
        return rearrange(x, "... h w (nh nw e) -> ... (h nh) (w nw) e", nh=self.h, nw=self.w)


class TokenSplit(nn.Module):
    def __init__(self, in_features, out_features, patch_size=(2, 2)):
        super().__init__()
        self.h = patch_size[0]
        self.w = patch_size[1]
        self.proj = apply_wd(Linear(in_features, out_features * self.h * self.w, bias=False))
        self.fac = nn.Parameter(torch.ones(1) * 0.5)
    
    def lerp(self, x, skip):
        return torch.lerp(skip, x, self.fac.to(x.dtype))

    def forward(self, x, skip):
        x = self.proj(x)
        x = rearrange(x, "... h w (nh nw e) -> ... (h nh) (w nw) e", nh=self.h, nw=self.w)
        return self.lerp(x, skip)


class Identity(nn.Identity):
    def forward(self, input, *args, **kwargs):
        return super().forward(input)


def zero_module(module: nn.Module):
    """
    Zero out the parameters of a module and return it.
    """
    for p in module.parameters():
        p.detach().zero_()
    return module


def bmm(a, b):
    return torch.einsum('b...i,bi...o->b...o', a, b)


#################################################################################################
# Copyright (c) 2022-2024 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
from typing import Optional
import torch
from torch import nn, Tensor
from torch.nn.functional import pad
from torch.nn.init import trunc_normal_
import natten
try:
    from natten.functional import na2d_qk_with_bias
except: pass
from einops import rearrange
import torch
from torch import nn
import torch._dynamo
from .. import flops
from .basic_core import scale_for_cosine_sim, checkpoint
from ..modules import AdaRMSNorm, Linear, AxialRoPE, FeedForwardBlock, apply_wd, zero_init, apply_rotary_emb_


def _pad_if_small(x, window_size, height, width):
    x = x.permute(0, 2, 3, 1)
    # Pad if the input is small than the minimum supported size
    padded_height, padded_width = height, width
    padding_h = padding_w = 0
    if height < window_size or width < window_size:
        padding_h = max(0, window_size - padded_height)
        padding_w = max(0, window_size - padded_width)
        x = pad(x, (0, 0, 0, padding_w, 0, padding_h))
        _, padded_height, padded_width, _ = x.shape
    return x.permute(0, 3, 1, 2), padded_height, padded_width


class NeighborhoodConvAttention(nn.Module):
    """
    Neighborhood Attention 2D Module
    """

    def __init__(
        self,
        dim_channels: int,
        dim_head: int | None = None,
        num_heads: int | None = None,
        kernel_size: int = 7,
        dilation: int = 1,
        bias: bool = True,
        qk_scale: Optional[float] = None,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        num_groups: int | None = 32
    ):
        super().__init__()
        self.group_norm = nn.Identity() if num_groups is None or num_groups == 0 else nn.GroupNorm(num_groups, dim_channels)
        
        if num_heads is None:
            num_heads = 1
            if dim_head is None:
                dim_head = dim_channels
        else:
            if dim_head is None:
                dim_head = dim_channels // num_heads
        dim_attn = dim_head * num_heads
        
        self.num_heads = num_heads
        self.dim_head = dim_head
        self.scale = qk_scale or self.dim_head ** -0.5
        
        self.proj_q = nn.Conv2d(dim_channels, dim_attn, 1, stride=1, padding=0)
        self.proj_k = nn.Conv2d(dim_channels, dim_attn, 1, stride=1, padding=0)
        self.proj_v = nn.Conv2d(dim_channels, dim_attn, 1, stride=1, padding=0)
        self.proj = nn.Conv2d(dim_attn, dim_channels, 1, stride=1, padding=0)
        
        assert (
            kernel_size > 1 and kernel_size % 2 == 1
        ), f"Kernel size must be an odd number greater than 1, got {kernel_size}."
        self.kernel_size = kernel_size
        assert (
            dilation is None or dilation >= 1
        ), f"Dilation must be greater than or equal to 1, got {dilation}."
        
        self.dilation = dilation or 1
        self.window_size = self.kernel_size * self.dilation

        if bias:
            self.rpb = nn.Parameter(
                torch.zeros(num_heads, (2 * kernel_size - 1), (2 * kernel_size - 1))
            )
            trunc_normal_(self.rpb, std=0.02, mean=0.0, a=-2.0, b=2.0)
        else:
            self.register_parameter("rpb", None)
        
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: Tensor) -> Tensor:
        x = self.group_norm(x)
        
        batch_size, _, height, width= x.shape
        x, padded_height, padded_width = _pad_if_small(x, self.window_size, height, width)
        
        q = self.proj_q(x).view(batch_size, self.num_heads, self.dim_head, padded_height, padded_width).permute(0, 1, 3, 4, 2)
        k = self.proj_k(x).view(batch_size, self.num_heads, self.dim_head, padded_height, padded_width).permute(0, 1, 3, 4, 2)
        v = self.proj_v(x).view(batch_size, self.num_heads, self.dim_head, padded_height, padded_width).permute(0, 1, 3, 4, 2)
        
        q = q * self.scale
        attn = na2d_qk_with_bias(q, k, self.rpb, self.kernel_size, self.dilation)
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        x = natten.functional.na2d_av(attn, v, self.kernel_size, self.dilation)
        x = x.reshape(batch_size, padded_height, padded_width, -1)

        # Remove padding, if added any
        if padded_height or padded_width:
            x = x[:, :height, :width, :]
        x = x.permute(0, 3, 1, 2)
        
        return self.proj_drop(self.proj(x))

    def extra_repr(self) -> str:
        return (
            f"head_dim={self.head_dim}, num_heads={self.num_heads}, "
            + f"kernel_size={self.kernel_size}, dilation={self.dilation}, "
            + f"has_bias={self.rpb is not None}"
        )


class NeighborhoodSelfAttentionBlock(nn.Module):
    def __init__(self, d_model, d_head, cond_features = None, kernel_size = 7, dropout = 0.0):
        super().__init__()
        self.d_head = d_head
        self.n_heads = d_model // d_head
        self.kernel_size = kernel_size
        self.norm = AdaRMSNorm(d_model, cond_features)
        self.qkv_proj = apply_wd(Linear(d_model, d_model * 3, bias=False))
        self.scale = nn.Parameter(torch.full([self.n_heads], 10.0))
        self.pos_emb = AxialRoPE(d_head // 2, self.n_heads)
        self.dropout = nn.Dropout(dropout)
        self.out_proj = apply_wd(zero_init(Linear(d_model, d_model, bias=False)))

    def extra_repr(self):
        return f"d_head={self.d_head}, kernel_size={self.kernel_size}"

    def forward(self, x, pos, cond = None):
        skip = x
        x = self.norm(x, cond)
        qkv = self.qkv_proj(x)
        if natten.has_fused_na():
            q, k, v = rearrange(qkv, "n h w (t nh e) -> t n h w nh e", t=3, e=self.d_head)
            q, k = scale_for_cosine_sim(q, k, self.scale[:, None], 1e-6)
            theta = self.pos_emb(pos)
            q = apply_rotary_emb_(q, theta)
            k = apply_rotary_emb_(k, theta)
            flops.op(flops.op_natten, q.shape, k.shape, v.shape, self.kernel_size)
            x = natten.functional.na2d(q, k, v, self.kernel_size, scale=1.0)
            x = rearrange(x, "n h w nh e -> n h w (nh e)")
        else:
            q, k, v = rearrange(qkv, "n h w (t nh e) -> t n nh h w e", t=3, e=self.d_head)
            q, k = scale_for_cosine_sim(q, k, self.scale[:, None, None, None], 1e-6)
            theta = self.pos_emb(pos).movedim(-2, -4)
            q = apply_rotary_emb_(q, theta)
            k = apply_rotary_emb_(k, theta)
            flops.op(flops.op_natten, q.shape, k.shape, v.shape, self.kernel_size)
            qk = natten.functional.na2d_qk(q, k, self.kernel_size)
            a = torch.softmax(qk, dim=-1).to(v.dtype)
            x = natten.functional.na2d_av(a, v, self.kernel_size)
            x = rearrange(x, "n nh h w e -> n h w (nh e)")
        x = self.dropout(x)
        x = self.out_proj(x)
        return x + skip


class NeighborhoodTransformerLayer(nn.Module):
    def __init__(self, d_model, d_ff, d_head, cond_features = None, kernel_size = 7, dropout = 0.0):
        super().__init__()
        self.self_attn = NeighborhoodSelfAttentionBlock(d_model, d_head, cond_features, kernel_size, dropout=dropout)
        self.ff = FeedForwardBlock(d_model, d_ff, cond_features, dropout=dropout)

    def forward(self, x, pos, cond = None):
        x = checkpoint(self.self_attn, x, pos, cond)
        x = checkpoint(self.ff, x, cond)
        return x


from einops import rearrange
import torch
from torch import nn
import torch._dynamo
from torch.nn import functional as F
from .. import flags, flops
from .basic_core import  scale_for_cosine_sim, scale_for_cosine_sim_qkv, checkpoint
from .. import flops, flags
from ..modules import AdaRMSNorm, RMSNorm, Linear, AxialRoPE, FeedForwardBlock, apply_wd, zero_init, apply_rotary_emb_
try:
    import flash_attn
except ImportError:
    flash_attn = None

def use_flash_2(x):
    if not flags.get_use_flash_attention_2():
        return False
    if flash_attn is None:
        return False
    if x.device.type != "cuda":
        return False
    if x.dtype not in (torch.float16, torch.bfloat16):
        return False
    return True


class SelfAttentionBlock(nn.Module):
    def __init__(self, d_model, d_head, cond_features = None, dropout=0.0):
        super().__init__()
        self.d_head = d_head
        self.n_heads = d_model // d_head
        if cond_features is None:
            self.norm = RMSNorm(d_model)
        else:
            self.norm = AdaRMSNorm(d_model, cond_features)
        self.qkv_proj = apply_wd(Linear(d_model, d_model * 3, bias=False))
        self.scale = nn.Parameter(torch.full([self.n_heads], 10.0))
        self.pos_emb = AxialRoPE(d_head // 2, self.n_heads)
        self.dropout = nn.Dropout(dropout)
        self.out_proj = apply_wd(zero_init(Linear(d_model, d_model, bias=False)))

    def extra_repr(self):
        return f"d_head={self.d_head},"

    def forward(self, x, pos, cond = None):
        skip = x
        x = self.norm(x, cond)
        qkv = self.qkv_proj(x)
        pos = rearrange(pos, "... h w e -> ... (h w) e").to(qkv.dtype)
        theta = self.pos_emb(pos)
        if use_flash_2(qkv):
            qkv = rearrange(qkv, "n h w (t nh e) -> n (h w) t nh e", t=3, e=self.d_head)
            qkv = scale_for_cosine_sim_qkv(qkv, self.scale, 1e-6)
            theta = torch.stack((theta, theta, torch.zeros_like(theta)), dim=-3)
            qkv = apply_rotary_emb_(qkv, theta)
            flops_shape = qkv.shape[-5], qkv.shape[-2], qkv.shape[-4], qkv.shape[-1]
            flops.op(flops.op_attention, flops_shape, flops_shape, flops_shape)
            x = flash_attn.flash_attn_qkvpacked_func(qkv, softmax_scale=1.0)
            x = rearrange(x, "n (h w) nh e -> n h w (nh e)", h=skip.shape[-3], w=skip.shape[-2])
        else:
            q, k, v = rearrange(qkv, "n h w (t nh e) -> t n nh (h w) e", t=3, e=self.d_head)
            q, k = scale_for_cosine_sim(q, k, self.scale[:, None, None], 1e-6)
            theta = theta.movedim(-2, -3)
            q = apply_rotary_emb_(q, theta)
            k = apply_rotary_emb_(k, theta)
            flops.op(flops.op_attention, q.shape, k.shape, v.shape)
            x = F.scaled_dot_product_attention(q, k, v, scale=1.0)
            x = rearrange(x, "n nh (h w) e -> n h w (nh e)", h=skip.shape[-3], w=skip.shape[-2])
        x = self.dropout(x)
        x = self.out_proj(x)
        return x + skip
    

class GlobalTransformerLayer(nn.Module):
    def __init__(self, d_model, d_ff, d_head, cond_features = None, dropout=0.0):
        super().__init__()
        self.self_attn = SelfAttentionBlock(d_model, d_head, cond_features, dropout=dropout)
        self.ff = FeedForwardBlock(d_model, d_ff, cond_features, dropout=dropout)

    def forward(self, x, pos, cond = None):
        x = checkpoint(self.self_attn, x, pos, cond)
        x = checkpoint(self.ff, x, cond)
        return x


class NoAttentionTransformerLayer(nn.Module):
    def __init__(self, d_model, d_ff, cond_features, dropout=0.0):
        super().__init__()
        self.ff = FeedForwardBlock(d_model, d_ff, cond_features, dropout=dropout)

    def forward(self, x, pos, cond):
        x = checkpoint(self.ff, x, cond)
        return x


import torch
from functools import reduce
from .. import flags


@flags.compile_wrap
def scale_for_cosine_sim(q, k, scale, eps):
    dtype = reduce(torch.promote_types, (q.dtype, k.dtype, scale.dtype, torch.float32))
    sum_sq_q = torch.sum(q.to(dtype)**2, dim=-1, keepdim=True)
    sum_sq_k = torch.sum(k.to(dtype)**2, dim=-1, keepdim=True)
    sqrt_scale = torch.sqrt(scale.to(dtype))
    scale_q = sqrt_scale * torch.rsqrt(sum_sq_q + eps)
    scale_k = sqrt_scale * torch.rsqrt(sum_sq_k + eps)
    return q * scale_q.to(q.dtype), k * scale_k.to(k.dtype)


@flags.compile_wrap
def scale_for_cosine_sim_qkv(qkv, scale, eps):
    q, k, v = qkv.unbind(2)
    q, k = scale_for_cosine_sim(q, k, scale[:, None], eps)
    return torch.stack((q, k, v), dim=2)


def checkpoint(function, *args, **kwargs):
    if flags.get_checkpointing():
        kwargs.setdefault("use_reentrant", True)
        return torch.utils.checkpoint.checkpoint(function, *args, **kwargs)
    else:
        return function(*args, **kwargs)


from contextlib import contextmanager
import math
import threading


state = threading.local()
state.flop_counter = None


@contextmanager
def flop_counter(enable=True):
    try:
        old_flop_counter = state.flop_counter
        state.flop_counter = FlopCounter() if enable else None
        yield state.flop_counter
    finally:
        state.flop_counter = old_flop_counter


class FlopCounter:
    def __init__(self):
        self.ops = []

    def op(self, op, *args, **kwargs):
        self.ops.append((op, args, kwargs))

    @property
    def flops(self):
        flops = 0
        for op, args, kwargs in self.ops:
            flops += op(*args, **kwargs)
        return flops


def op(op, *args, **kwargs):
    if getattr(state, "flop_counter", None):
        state.flop_counter.op(op, *args, **kwargs)


def op_linear(x, weight):
    return math.prod(x) * weight[0]


def op_attention(q, k, v):
    *b, s_q, d_q = q
    *b, s_k, d_k = k
    *b, s_v, d_v = v
    return math.prod(b) * s_q * s_k * (d_q + d_v)


def op_natten(q, k, v, kernel_size):
    *q_rest, d_q = q
    *_, d_v = v
    return math.prod(q_rest) * (d_q + d_v) * kernel_size**2



import inspect
from typing import Iterable


class ArgsContext:
    def __init__(self, **kwargs):
        self.__args_context = kwargs
    
    def __getitem__(self, name):
        return self.__args_context[name]
    
    def __getattr__(self, name):
        return self.__args_context[name]
    
    @classmethod
    def __class_getitem__(cls, typename):
        if isinstance(typename, tuple):
            typename, exception_types = typename
            if not isinstance(exception_types, Iterable):
                exception_types = [exception_types]
        else:
            exception_types = []
        class ArgsContext:
            def __init__(self, *args, **kwargs):
                signature = inspect.signature(typename.__init__)
                parameters = signature.parameters
                parameters = [p for p in signature.parameters.values() if p.name != 'self']
                signature = signature.replace(parameters=parameters)
                bound_arguments = signature.bind(*args, **kwargs)
                bound_arguments.apply_defaults()
                for name, value in bound_arguments.arguments.items():
                    if not name in exception_types:
                        setattr(self, name, value)
        return ArgsContext


from . import neighborhood_attention
from . import self_attention
from . import shifted_self_attention
from .neighborhood_attention import NeighborhoodConvAttention, NeighborhoodSelfAttentionBlock, NeighborhoodTransformerLayer
from .self_attention import SelfAttentionBlock, GlobalTransformerLayer
from .shifted_self_attention import ShiftedWindowSelfAttentionBlock, ShiftedWindowTransformerLayer


from contextlib import contextmanager
from functools import update_wrapper
import os
import threading
import torch


def get_use_compile():
    return os.environ.get("K_DIFFUSION_USE_COMPILE", "1") == "1"


def get_use_flash_attention_2():
    return os.environ.get("K_DIFFUSION_USE_FLASH_2", "1") == "1"


state = threading.local()
state.checkpointing = False


@contextmanager
def checkpointing(enable=True):
    try:
        old_checkpointing, state.checkpointing = state.checkpointing, enable
        yield
    finally:
        state.checkpointing = old_checkpointing


def get_checkpointing():
    return getattr(state, "checkpointing", False)


class compile_wrap:
    def __init__(self, function, *args, **kwargs):
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self._compiled_function = None
        update_wrapper(self, function)

    @property
    def compiled_function(self):
        if self._compiled_function is not None:
            return self._compiled_function
        if get_use_compile():
            try:
                self._compiled_function = torch.compile(self.function, *self.args, **self.kwargs)
            except RuntimeError:
                self._compiled_function = self.function
        else:
            self._compiled_function = self.function
        return self._compiled_function

    def __call__(self, *args, **kwargs):
        return self.compiled_function(*args, **kwargs)



from einops import rearrange
import torch
from torch import nn
import torch._dynamo
from torch.nn import functional as F
from functools import lru_cache
from .. import flops
from .basic_core import scale_for_cosine_sim, checkpoint
from ..modules import AdaRMSNorm, Linear, AxialRoPE, FeedForwardBlock, apply_wd, zero_init, apply_rotary_emb_


def window(window_size, x):
    *b, h, w, c = x.shape
    x = torch.reshape(
        x,
        (*b, h // window_size, window_size, w // window_size, window_size, c),
    )
    x = torch.permute(
        x,
        (*range(len(b)), -5, -3, -4, -2, -1),
    )
    return x


def unwindow(x):
    *b, h, w, wh, ww, c = x.shape
    x = torch.permute(x, (*range(len(b)), -5, -3, -4, -2, -1))
    x = torch.reshape(x, (*b, h * wh, w * ww, c))
    return x


def shifted_window(window_size, window_shift, x):
    x = torch.roll(x, shifts=(window_shift, window_shift), dims=(-2, -3))
    windows = window(window_size, x)
    return windows


def shifted_unwindow(window_shift, x):
    x = unwindow(x)
    x = torch.roll(x, shifts=(-window_shift, -window_shift), dims=(-2, -3))
    return x


@lru_cache
def make_shifted_window_masks(n_h_w, n_w_w, w_h, w_w, shift, device=None):
    ph_coords = torch.arange(n_h_w, device=device)
    pw_coords = torch.arange(n_w_w, device=device)
    h_coords = torch.arange(w_h, device=device)
    w_coords = torch.arange(w_w, device=device)
    patch_h, patch_w, q_h, q_w, k_h, k_w = torch.meshgrid(
        ph_coords,
        pw_coords,
        h_coords,
        w_coords,
        h_coords,
        w_coords,
        indexing="ij",
    )
    is_top_patch = patch_h == 0
    is_left_patch = patch_w == 0
    q_above_shift = q_h < shift
    k_above_shift = k_h < shift
    q_left_of_shift = q_w < shift
    k_left_of_shift = k_w < shift
    m_corner = (
        is_left_patch
        & is_top_patch
        & (q_left_of_shift == k_left_of_shift)
        & (q_above_shift == k_above_shift)
    )
    m_left = is_left_patch & ~is_top_patch & (q_left_of_shift == k_left_of_shift)
    m_top = ~is_left_patch & is_top_patch & (q_above_shift == k_above_shift)
    m_rest = ~is_left_patch & ~is_top_patch
    m = m_corner | m_left | m_top | m_rest
    return m


def apply_window_attention(window_size, window_shift, q, k, v, scale=None):
    # prep windows and masks
    q_windows = shifted_window(window_size, window_shift, q)
    k_windows = shifted_window(window_size, window_shift, k)
    v_windows = shifted_window(window_size, window_shift, v)
    b, heads, h, w, wh, ww, d_head = q_windows.shape
    mask = make_shifted_window_masks(h, w, wh, ww, window_shift, device=q.device)
    q_seqs = torch.reshape(q_windows, (b, heads, h, w, wh * ww, d_head))
    k_seqs = torch.reshape(k_windows, (b, heads, h, w, wh * ww, d_head))
    v_seqs = torch.reshape(v_windows, (b, heads, h, w, wh * ww, d_head))
    mask = torch.reshape(mask, (h, w, wh * ww, wh * ww))

    # do the attention here
    flops.op(flops.op_attention, q_seqs.shape, k_seqs.shape, v_seqs.shape)
    qkv = F.scaled_dot_product_attention(q_seqs, k_seqs, v_seqs, mask, scale=scale)

    # unwindow
    qkv = torch.reshape(qkv, (b, heads, h, w, wh, ww, d_head))
    return shifted_unwindow(window_shift, qkv)


class ShiftedWindowSelfAttentionBlock(nn.Module):
    def __init__(self, d_model, d_head, cond_features, window_size, window_shift, dropout=0.0):
        super().__init__()
        self.d_head = d_head
        self.n_heads = d_model // d_head
        self.window_size = window_size
        self.window_shift = window_shift
        self.norm = AdaRMSNorm(d_model, cond_features)
        self.qkv_proj = apply_wd(Linear(d_model, d_model * 3, bias=False))
        self.scale = nn.Parameter(torch.full([self.n_heads], 10.0))
        self.pos_emb = AxialRoPE(d_head // 2, self.n_heads)
        self.dropout = nn.Dropout(dropout)
        self.out_proj = apply_wd(zero_init(Linear(d_model, d_model, bias=False)))

    def extra_repr(self):
        return f"d_head={self.d_head}, window_size={self.window_size}, window_shift={self.window_shift}"

    def forward(self, x, pos, cond):
        skip = x
        x = self.norm(x, cond)
        qkv = self.qkv_proj(x)
        q, k, v = rearrange(qkv, "n h w (t nh e) -> t n nh h w e", t=3, e=self.d_head)
        q, k = scale_for_cosine_sim(q, k, self.scale[:, None, None, None], 1e-6)
        theta = self.pos_emb(pos).movedim(-2, -4)
        q = apply_rotary_emb_(q, theta)
        k = apply_rotary_emb_(k, theta)
        x = apply_window_attention(self.window_size, self.window_shift, q, k, v, scale=1.0)
        x = rearrange(x, "n nh h w e -> n h w (nh e)")
        x = self.dropout(x)
        x = self.out_proj(x)
        return x + skip


class ShiftedWindowTransformerLayer(nn.Module):
    def __init__(self, d_model, d_ff, d_head, cond_features, window_size, index, dropout=0.0):
        super().__init__()
        window_shift = window_size // 2 if index % 2 == 1 else 0
        self.self_attn = ShiftedWindowSelfAttentionBlock(d_model, d_head, cond_features, window_size, window_shift, dropout=dropout)
        self.ff = FeedForwardBlock(d_model, d_ff, cond_features, dropout=dropout)

    def forward(self, x, pos, cond):
        x = checkpoint(self.self_attn, x, pos, cond)
        x = checkpoint(self.ff, x, cond)
        return x